@@ -30,6 +30,7 @@ from .AdcMatrix import AdcMatrix, AdcMatrixlike
 from .AdcMethod import AdcMethod
 from .ExcitedStates import ExcitedStates
 from .ReferenceState import ReferenceState as adcc_ReferenceState
+
 from .solver.davidson import jacobi_davidson
 from .solver.explicit_symmetrisation import (IndexSpinSymmetrisation,
                                              IndexSymmetrisation)
@@ -81,7 +82,7 @@ def run_adc(data_or_matrix, n_states=None, kind="any", conv_tol=None,
 
     conv_tol : float, optional
         Convergence tolerance to employ in the iterative solver for obtaining
-        the ADC vectors (default: `1e-6` or SCF tolerance / 100,
+        the ADC vectors (default: `1e-6` or 10 * SCF tolerance,
         whatever is larger)
 
     solver_method : str, optional
@@ -333,12 +334,12 @@ def diagonalise_adcmatrix(matrix, n_states, kind, solver_method="davidson",
 
     # Determine default ADC convergence tolerance
     if conv_tol is None:
-        conv_tol = max(reference_state.conv_tol / 100, 1e-6)
+        conv_tol = max(10 * reference_state.conv_tol, 1e-6)
     if reference_state.conv_tol > conv_tol:
         raise ValueError(
-            f"Convergence tolerance of SCF results "
-            "(== {reference_state.conv_tol}) needs to be lower than ADC "
-            "convergence tolerance parameter conv_tol (== {conv_tol})."
+            "Convergence tolerance of SCF results "
+            f"(== {reference_state.conv_tol}) needs to be lower than ADC "
+            f"convergence tolerance parameter conv_tol (== {conv_tol})."
         )
 
     # Determine explicit_symmetrisation
